# experiments

using Random: MersenneTwister
using LinearAlgebra: Diagonal, I, dot, norm, mul!, diagm
using Statistics: median

using CSV
using Colors: HSV
using DataFrames
using Distributed
@everywhere using DualTVDD
@everywhere using DualTVDD:
    DualTVL1ROFOpProblem,
    DualTVDDAlgorithm, ChambolleAlgorithm, NStepAlgorithm, ProjGradAlgorithm,# TVNewtonAlgorithm,
    SurrogateAlgorithm,
    DualTVDDSurrogateAlgorithm,
    init, step!, fetch, fetch_u, gradient, divergence, normB, energy, residual
using DualTVDD: fetch_u
using FFTW
using FileIO
using ImageIO
using IterativeSolvers
using LinearMaps
using OpticalFlowUtils
using Optim
using Plots
using StaticKernels: Kernel, ExtensionReplicate, ExtensionNothing, extend

include("util.jl")
isdefined(Main, :Revise) && Revise.track(joinpath(@__DIR__, "util.jl"))
include("wavelet.jl")
isdefined(Main, :Revise) && Revise.track(joinpath(@__DIR__, "wavelet.jl"))

using .Util

"""
    logfilter(dt; a=20)

Reduce dataset such that the resulting dataset would have approximately
equidistant datapoints on a log-scale. `a` controls density.
"""
logfilter(dt; a=20) = filter(:k => k -> round(a*log(k+1)) - round(a*log(k)) > 0, dt)

"""
    halve(img)

Scale `img` to half size using box-avaraging.
"""
function halve(img)
    all(iseven.(size(img))) || throw(ArgumentError("uneven size"))

    res = similar(img, size(img) .÷ 2)
    fill!(res, zero(eltype(res)))

    for i in CartesianIndices(img)
        j = CartesianIndex((Tuple(i) .+ 1) .÷ 2)
        res[j] += img[i]
    end
    res ./= 2 ^ ndims(img)

    return res
end

# Problem Definitions

function DenoiseProblem(img::AbstractArray{T,d}; λ, β = 0.) where {T,d}
    B = (1 + β) * I
    g = img

    return DualTVL1ROFOpProblem(g, B, λ)
end

function InpaintProblem(img::AbstractArray{T,d}, imgmask::AbstractArray{Bool,d}; λ, β) where {T,d}
    pwop = imgmask .+ β .* Ref(I)
    B = Diagonal(vec(inv.(pwop)))

    g = imgmask .* img
    return DualTVL1ROFOpProblem(g, B, λ)
end

function OptFlowProblem(f0::AbstractArray{T,d}, f1::AbstractArray{T,d}; λ, β) where {T,d}
    size(f0) == size(f1) || throw(ArgumentError("mismatched sizes"))

    ∇ = Kernel{ntuple(_->0:1, d)}(gradient)
    ∇f0 = map(∇, extend(f0, ExtensionReplicate()))

    Bval = inv.(∇f0 .* adjoint.(∇f0) .+ β .* Ref(I))
    B = Diagonal(vec(Bval))

    g = .-(f1 .- f0) .* ∇f0

    return DualTVL1ROFOpProblem(g, B, λ)
end

function DFTProblem(img::AbstractArray{T,d}; λ, β=0.) where {T,d}
    T <: Complex || throw(ArgumentError("need `<:Complex` eltype for frequency data"))
    n = length(img)
    B = 1 / (n + β) * I
    g = real.(n * ifft(img))

    return DualTVL1ROFOpProblem(g, B, λ)
end

function GlobalProblem(img::AbstractArray{T,d}, A; λ, β) where {T,d}
    n = length(img)

    B = LinearMap{Float64}(
        b -> cg(A'*A + β * I, b; maxiter=10),
        n, n; issymmetric=true, isposdef=β>0)

    #B = inv(A'*A + β*I)
    #B = (B' + B) / 2
    g = reshape(A' * A * vec(img), size(img))

    return DualTVL1ROFOpProblem(g, B, λ)
end

# Util

toimg(arr) = Gray.(clamp.(arr, 0., 1.))
loadimg(x) = Float64.(load(x))
saveimg(io, x) = save(io, toimg(x))

function run_algs(dataf, algs; maxiters)
    flatten(x; kwargs...) =
        (;kwargs..., (Symbol(a, "_", c) => d for (a, b) in pairs(x) for (c, d) in pairs(b))...)

    df = DataFrame()
    states = map(init, algs)
    push!(df, flatten(map(dataf, states); k=0))

    # log-skipping parameter
    a = 20
    for k in 1:maxiters
        round(a*log(k+1)) - round(a*log(k)) == 0 && continue
        push!(df.k, k)
    end
    for (algname, st) in pairs(states)
        for k in 1:maxiters
            step!(st)
            round(a*log(k+1)) - round(a*log(k)) == 0 && continue
            stepdata = dataf(st)
            for (dataname, data) in pairs(stepdata)
                push!(getproperty(df, Symbol(algname, "_", dataname)), data)
            end
        end
    end

    return df, states
end

comparison_algs(prob; τ, M, overlap, ninner) = (
    glob = NStepAlgorithm(ChambolleAlgorithm(prob; τ), ninner),
    ddseq = DualTVDDAlgorithm(prob; M, overlap, ninner, parallel=false, σ=1., subalg=x->ChambolleAlgorithm(x; τ)),
    ddseqav = DualTVDDAlgorithm(prob; M, overlap, ninner, parallel=false, σ=0.25, subalg=x->ChambolleAlgorithm(x; τ)),
    ddpar = DualTVDDAlgorithm(prob; M, overlap, ninner, parallel=true, σ=0.25, subalg=x->ChambolleAlgorithm(x; τ)))

function calc_energymin(prob, alg; maxiters)
    st = init(alg)
    for _ in 1:maxiters
        step!(st)
    end
    return energy(fetch(st), prob)
end

# Experiment Suite

function compare_algs(prob; τ, M, overlap, maxiters, ninner)
    algs = comparison_algs(prob; τ, M, overlap, ninner)

    df, states = run_algs(algs; maxiters) do state
        return (
            energy = energy(fetch(state), prob),
            residual = residual(fetch(state), prob))
    end

    #filter!(:k => k -> k < maxiters / 10, df)

    # subtract minimal energy for logarithmic plot
    energymin = calc_energymin(prob, algs.glob; maxiters=maxiters * 10)
    for x in names(df)
        contains(x, r"_energy$") || continue
        df[!,Symbol(x)] .-= energymin
    end

    return df, states, energymin
end

function run_while(fwhile, alg)
    st = init(alg)
    k = 0
    while fwhile(st, k)
        k += 1
        step!(st)
    end
    close(st)
    return st
end

function experiment_convergence_test(ctx)
    #fo = loadimg(joinpath(ctx.indir, "input_original.png"))

    #λ = 0.1
    #β = 0.
    #maxiters = 10
    #ninner = 10

    #M = (2,2)
    #overlap = (5,5)

    #noise_σ = 0.1

    #seed = 42
    #rng = MersenneTwister(seed)
    #f = clamp.(fo .+ noise_σ * randn(rng, Float64, size(fo)), 0., 1.)
    #saveimg(joinpath(ctx.outdir, "input_noisy.png"), f)

    #prob = DenoiseProblem(f; λ, β)

    fo = f0 = loadimg(joinpath(ctx.indir, "input_f0.png"))
    f1 = loadimg(joinpath(ctx.indir, "input_f1.png"))

    λ = 0.002
    β = 0.01

    maxiters = 100
    ninner = 1000

    M = (4,4)
    overlap = (5,5)

    prob = OptFlowProblem(f0, f1; λ, β)

    algs = (;
        #glob = ChambolleAlgorithm(prob),
        #surrogate = SurrogateAlgorithm(prob; ninner=100, τ=2*normB(prob)+eps(normB(prob)),
        #                               subalg=x->ChambolleAlgorithm(x)),

        surrogate_ddseq = SurrogateAlgorithm(prob; ninner=1, τ = normB(prob) + eps(normB(prob)),
            subalg = x -> DualTVDDAlgorithm(x; M, overlap, ninner, parallel = false, σ = 1.,
                subalg = x -> ChambolleAlgorithm(x))),
        ddseq_surrogate = DualTVDDAlgorithm(prob; M, overlap, ninner, parallel = false, σ = 1.,
            subalg = x -> SurrogateAlgorithm(x; ninner=1, τ = normB(x) + eps(normB(x)),
                subalg = x -> ChambolleAlgorithm(x))),
        #ddseq = DualTVDDAlgorithm(prob; M, overlap, ninner, parallel = false, σ = 1.,
        #    subalg = x -> ChambolleAlgorithm(x)),
       )

    #algs = comparison_algs(prob; τ, M, overlap, ninner)

    df, states = run_algs(algs; maxiters) do state
        return (
            energy = energy(fetch(state), prob),
            residual = residual(fetch(state), prob))
    end
    display(df)

    CSV.write(joinpath(ctx.outdir, "energies.csv"), df)
    #saveimg(joinpath(ctx.outdir, "output_glob.png"), fetch_u(states.glob))
    savedata(joinpath(ctx.outdir, "data.tex");
        lambda=λ, beta=β, maxiters,
        width=size(fo, 2), height=size(fo, 1))
end

function experiment_scaling_opticalflow(ctx)
    fo = f0 = loadimg(joinpath(ctx.indir, "frame10.png"))
    f1 = loadimg(joinpath(ctx.indir, "frame11.png"))
    d = 2

    λ = 0.01
    β = 0.001

    ninner = 300
    Mdir = 2 * floor(Int, sqrt(nworkers())) # to have enough workers available
    M = (Mdir, Mdir)
    overlap = (5, 5)

    stopenergy = 130.
    ntimings = 3

    prob = OptFlowProblem(f0, f1; λ, β)

    galg = ChambolleAlgorithm(prob)
    alg_ddseq(workers) = DualTVDDAlgorithm(prob; workers, M, overlap, ninner, parallel = false, σ = 1., subalg = x -> ChambolleAlgorithm(x))
    alg_ddpar(workers) = DualTVDDAlgorithm(prob; workers, M, overlap, ninner, parallel = true, σ = 0.25, subalg = x -> ChambolleAlgorithm(x))

    function timeit(alg)
        # warmup for precompilation
        run_while((_, _) -> false, alg)

        times = map(1:ntimings) do _
            return @elapsed run_while(alg) do st, k
                return energy(fetch(st), prob) > stopenergy
            end
        end
        return minimum(times)
    end

    tg = timeit(galg)

    nparallel = prod(M) ÷ 2^2
    ws = workers()
    @assert nparallel <= length(ws)
    df = DataFrame()
    for nw in 1:nparallel
        nparallel % nw == 0 || continue
        push!(df, (
            nworkers = nw,
            time_ddseq = timeit(alg_ddseq(ws[1:nw])),
            time_ddpar = timeit(alg_ddpar(ws[1:nw]))
        ))
    end

    display(df)

    CSV.write(joinpath(ctx.outdir, "timings.csv"), df)
    savedata(joinpath(ctx.outdir, "data.tex");
        lambda=λ, beta=β, Mdir, M=prod(M),
        ntimings, stopenergy, ninner,
        width=size(fo, 2), height=size(fo, 1))
end

experiment_surrogate_outerinner(ctx) =
    _experiment_surrogate_outerinner(ctx, false)
experiment_surrogate_outerinner_ref(ctx) =
    _experiment_surrogate_outerinner(ctx, true)

function _experiment_surrogate_outerinner(ctx, ref)
    fo = f0 = loadimg(joinpath(ctx.indir, "frame10.png"))
    f1 = loadimg(joinpath(ctx.indir, "frame11.png"))

    λ = 0.01
    β = 0.001
    prob = OptFlowProblem(f0, f1; λ, β)

    maxiters = 100
    ninner = 50
    minenergy_maxiters = 10 * maxiters * ninner

    M = (4,4)
    overlap = (50,50)

    # set B = I for reference
    if ref
        prob.B.diag .= Ref(Matrix(I, 2, 2))
    end

    energymin = calc_energymin(prob, ChambolleAlgorithm(prob); maxiters=minenergy_maxiters)

    algs = (;
        surrogate_ddseq = SurrogateAlgorithm(prob; ninner=1, τ = normB(prob) + eps(normB(prob)),
            subalg = x -> DualTVDDAlgorithm(x; M, overlap, ninner, parallel = false, σ = 1.,
                subalg = x -> ChambolleAlgorithm(x))),
        #surrogate_ddpar = SurrogateAlgorithm(prob; ninner=1, τ = normB(prob) + eps(normB(prob)),
        #    subalg = x -> DualTVDDAlgorithm(x; M, overlap, ninner, parallel = true, σ = 0.25,
        #        subalg = x -> ChambolleAlgorithm(x))),
        ddseq_surrogate = DualTVDDAlgorithm(prob; M, overlap, ninner=1, parallel = false, σ = 1.,
            subalg = x -> SurrogateAlgorithm(x; ninner, τ = normB(x) + eps(normB(x)),
                subalg = x -> ChambolleAlgorithm(x))),
        #ddpar_surrogate = DualTVDDAlgorithm(prob; M, overlap, ninner=1, parallel = true, σ = 0.25,
        #    subalg = x -> SurrogateAlgorithm(x; ninner, τ = normB(x) + eps(normB(x)),
        #        subalg = x -> ChambolleAlgorithm(x))),
       )

    df, states = run_algs(algs; maxiters) do state
        return (
            energy = energy(fetch(state), prob) - energymin,
            #residual = residual(fetch(state), prob),
           )
    end
    #display(df)

    CSV.write(joinpath(ctx.outdir, "energies.csv"), df)
    #saveimg(joinpath(ctx.outdir, "output_glob.png"), fetch_u(states.glob))
    savedata(joinpath(ctx.outdir, "data.tex");
        lambda=λ, beta=β, minenergy_maxiters, maxiters, ninner, M, overlap, energymin,
        width=size(fo, 2), height=size(fo, 1))
end

function experiment_convergence_denoising(ctx)
    fo = loadimg(joinpath(ctx.indir, "input_original.png"))

    λ = 0.1
    β = 0.
    maxiters = 10000
    ninner = 100

    M = (2,2)
    overlap = (5,5)

    noise_σ = 0.1

    seed = 42
    rng = MersenneTwister(seed)
    f = clamp.(fo .+ noise_σ * randn(rng, Float64, size(fo)), 0., 1.)
    saveimg(joinpath(ctx.outdir, "input_noisy.png"), f)

    prob = DenoiseProblem(f; λ, β)

    τ = inv(8 * normB(prob))
    df, states, energymin = compare_algs(prob; τ, M, overlap, maxiters, ninner)

    CSV.write(joinpath(ctx.outdir, "energies.csv"), df)
    for alg in keys(states)
        saveimg(joinpath(ctx.outdir, "output_" * string(alg) * ".png"), fetch_u(states[alg]))
    end
    savedata(joinpath(ctx.outdir, "data.tex");
        lambda=λ, beta=β, tau=τ, maxiters, ninner, noise_sigma=noise_σ, energymin,
        width=size(fo, 2), height=size(fo, 1))
end

function experiment_convergence_inpainting(ctx)
    fo = loadimg(joinpath(ctx.indir, "input_original.png"))

    λ = 0.05
    β = 0.001
    maxiters = 10000
    ninner = 100

    M = (2,2)
    overlap = (5,5)


    seed = 42
    rng = MersenneTwister(seed)
    f_mask = rand(rng, Float64, size(fo)) .> 0.5

    prob = InpaintProblem(fo, f_mask; λ, β)
    saveimg(joinpath(ctx.outdir, "input_mask.png"), f_mask)
    saveimg(joinpath(ctx.outdir, "input_g.png"), prob.g)

    τ = inv(8 * normB(prob))
    df, states, energymin = compare_algs(prob; τ, M, overlap, maxiters, ninner)

    CSV.write(joinpath(ctx.outdir, "energies.csv"), df)
    for alg in keys(states)
        saveimg(joinpath(ctx.outdir, "output_" * string(alg) * ".png"), fetch_u(states[alg]))
    end
    savedata(joinpath(ctx.outdir, "data.tex");
        lambda=λ, beta=β, tau=τ, maxiters, ninner, energymin,
        width=size(fo, 2), height=size(fo, 1))
end

function experiment_convergence_opticalflow(ctx)
    f0 = loadimg(joinpath(ctx.indir, "input_f0.png"))
    f1 = loadimg(joinpath(ctx.indir, "input_f1.png"))

    λ = 0.002
    β = 0.001

    maxiters = 10000
    ninner = 100

    M = (2,2)
    overlap = (5,5)

    prob = OptFlowProblem(f0, f1; λ, β)

    τ = inv(8 * normB(prob))
    df, states, energymin = compare_algs(prob; τ, M, overlap, maxiters, ninner)

    CSV.write(joinpath(ctx.outdir, "energies.csv"), df)

    for alg in keys(states)
        u_nested = fetch_u(states[alg])
        u = reinterpret(reshape, Float64, u_nested)
        cflow = colorflow(u; maxflow=4.5/(584/48))
        #p = plot(cflow); display(p)
        save(joinpath(ctx.outdir, "flow_" * string(alg) * ".png"), cflow)
    end

    savedata(joinpath(ctx.outdir, "data.tex");
        lambda=λ, beta=β, tau=τ, maxiters, ninner, energymin,
        width=size(f0, 2), height=size(f0, 1))
end


function experiment_global_basic(ctx)
    f = loadimg(joinpath(ctx.indir, "input_original.png"))
    #f = loadimg(joinpath(ctx.indir, "frame10_small.png"))
    ax = axes(f)

    seed = 40
    rng = MersenneTwister(seed)

    λ = 0.02
    β = 1e-3

    #maxiters = 10000
    maxiters = 100
    ninner = 1000

    M = (4,4)
    overlap = (5,5)

    n = length(f)
    # masked wavelet coefficients
    wr = LinearIndices(size(f))[rand(rng, Float64, size(f)) .> 0.5]
    #wr = n÷4:n÷2

    A = LinearMap{Float64}(
        (y, x) -> (mdwt!(reshape(y, ax), reshape(x, ax)); y[wr] .= 0; y),
        (y, x) -> (x[wr] .= 0; imdwt!(reshape(y, ax), reshape(x, ax))),
        # no longer symmetric
        n, n, issymmetric=false, isposdef=false)

    #A = rand(rng, n, n) ./ n
    #A = diagm(rand(rng, n))

    #A = LinearMap{Float64}(x -> x, x -> x, n, n, issymmetric=true, isposdef=true)

    prob = GlobalProblem(f, A; λ, β)


    #τ = inv(8 * normB(prob))

    algref = ChambolleAlgorithm(prob, τ=β/8)
    st = init(algref)
    u0 = fetch_u(st)

    #alg = ChambolleAlgorithm(prob, τ=β/8)

    τ = 2/β
    alg = DualTVDDSurrogateAlgorithm(prob; M, overlap, ninner, τ, parallel=false)

    #alg = SurrogateAlgorithm(prob; ninner=1, τ,
    #    subalg = x -> DualTVDDAlgorithm(x; M, overlap, ninner, parallel = false, σ = 1.,
    #        subalg = x -> ChambolleAlgorithm(x)))

    #return alg
    #alg = DualTVDDAlgorithm(prob; M, overlap, ninner)

    #display(fetch(st))
    #display(fetch_u(st))
    #step!(st)
    #display(fetch(st))
    #display(fetch_u(st))
    #step!(st)
    #display(fetch(st))
    #display(fetch_u(st))
    #return

    df, states = run_algs([alg]; maxiters) do state
        return (
            energy = energy(fetch(state), prob),
            residual = residual(fetch(state), prob))
    end

    #energymin = calc_energymin(prob, algref; maxiters)
    #for x in names(df)
    #    contains(x, r"_energy$") || continue
    #    df[!,Symbol(x)] .-= energymin
    #end

    CSV.write(joinpath(ctx.outdir, "energies.csv"), df)

    u = fetch_u(states[1])
    #display(u)
    p = plot(plot(toimg(f)), plot(toimg(u0)), plot(toimg(u)))

    saveimg(joinpath(ctx.outdir, "input_orig.png"), f)
    saveimg(joinpath(ctx.outdir, "input_g.png"), prob.g)
    saveimg(joinpath(ctx.outdir, "output_u.png"), u)


    display(p)


    savedata(joinpath(ctx.outdir, "data.tex");
        lambda=λ, beta=β, tau=τ, maxiters, ninner,
        width=size(f, 2), height=size(f, 1))
    return prob
end
