using StaticArrays
using StaticKernels

"""
    <:Problem

The abstract type hierarchy specifies the problem model interface (problem,
available data, available oracles, solution form).
"""
abstract type Problem end

"""
    <:Algorithm{<:Problem}
    (<:Algorithm)(::Problem; params...)

An algorithm represents a fully specified iterative process of usually infite
or a-priori fixed but unknown finite length, producing iterates that
approximate the solution to some problem (see `::Problem`).

The hierarchy of abstract types `<:Algorithm` is based on the algorithm
interface (e.g. accepted parameters) and the specific numerical scheme.

A concrete type `<:Algorithm` represents an implementation of its supertype
algorithm. An instance is a complete specification of the algorithm with all
its inputs and should guarantee to produce a deterministic sequence of iterates
(randomized algorithms should use a seeded pseudorandom number generator).

An concrete algorithm type needs to implement at least

  - `state = init(::Algorithm)` allocates and initializes the algorithm state.
  - `state = step!(::State)` performs one iteration of the algorithm by
    updating `state`. This method should not allocate dynamic memory.
  - `solution = fetch(::State)` fetches the solution in a format specified by
    the problem. In some cases this can be a non-trivial operation if the
    algorithm uses a different internal representation.

Depending on its supertype algorithm interface additional methods might need to
be implemented.
"""
abstract type Algorithm{P<:Problem} end

"""
    <:State

The algorithm workspace. Together with the algorithm it contains sufficient
information to act as a checkpoint for continuing the algorithm at that point.
"""
abstract type State end

function init end
function Base.close(::State) end
function step! end
function fetch end

fetch_u(st) = recover_u(fetch(st), st.algorithm.problem)

@generated function gradient(w::StaticKernels.Window{<:Real,N}) where N
    i0 = ntuple(_->0, N)
    i1(k) = ntuple(i->Int(k==i), N)

    wi = (:(w[$(i1(k)...)] - w[$(i0...)]) for k in 1:N)
    return quote
        Base.@_inline_meta
        return @inbounds SVector($(wi...))
    end
end

@generated function gradient(w::StaticKernels.Window{S,N}) where {S<:SArray, N}
    i0 = ntuple(_->0, N)
    i1(k) = ntuple(i->Int(k==i), N)

    wi = (:( w[$(i1(k)...)][$j] - w[$(i0...)][$j] ) for k in 1:N for j in 1:length(S))
    return quote
        Base.@_inline_meta
        return @inbounds SArray{Tuple{$(size(S)...),N}}($(wi...))::
            SArray{Tuple{$(size(S)...),N},$(eltype(S)),$(ndims(S)+1),$(length(S)*N)}
    end
end

"""
    divergence(w::StaticKernels.Window)

Kernel function to compute the divergence of a vector field.
The window is expected to have eltype either:
  - `SVector{N,T}` and `N` dimensions or
  - `SMatrix{M,N,T}` and `N` dimensions.
"""
function divergence end

@generated function divergence(w::StaticKernels.Window{SVector{N,T},N}) where {N,T}
    i0 = ntuple(_->0, N)
    i1(k) = ntuple(i->Int(k==i), N)

    wi = (:((isnothing(w[$(i1(k)...)]) ? zero(T) : w[$(i0...)][$k]) -
            (isnothing(w[$((.-i1(k))...)]) ? zero(T) : w[$((.-i1(k))...)][$k])) for k in 1:N)
    return quote
        Base.@_inline_meta
        return @inbounds +($(wi...))
    end
end

@generated function divergence(w::StaticKernels.Window{S,N}) where {M,N,S<:SMatrix{M,N}}
    T = SVector{M,eltype(S)}
    sz = size(S)
    sz[end] == N || throw(ArgumentError("last eltype dimension does not match array dimensionality"))

    i0 = ntuple(_->0, N)
    i1(k) = ntuple(i->Int(k==i), N)

    slice(k) = (ntuple(_->:, ndims(S)-1)..., k)

    wi = (:((isnothing(w[$(i1(k)...)]) ? zero($T) : w[$(i0...)][$(slice(k)...)]) -
            (isnothing(w[$((.-i1(k))...)]) ? zero($T) : w[$((.-i1(k))...)][$(slice(k)...)])) for k in 1:N)
    return quote
        Base.@_inline_meta
        return @inbounds +($(wi...))
    end
end

function div_op(a::AbstractArray{<:StaticVector{N},N}) where N
    k = Kernel{ntuple(_->-1:1, ndims(a))}(k)
    ae = extend(a, StaticKernels.ExtensionNothing())
    return AppliedLinearKernel(k, ae)
end

function grad_op(a::AbstractArray)
    k = Kernel{ntuple(_->0:1, ndims(a))}(k)
    ae = extend(a, StaticKernels.ExtensionReplicate())
    return AppliedLinearKernel(k, ae)
end
